﻿using System;
using System.Runtime.InteropServices;

namespace AC.Input.Platforms.Windows
{
	/// <summary>
	/// http://www.activevb.de/cgi-bin/apiwiki/JOYINFOEX
	/// 
	/// http://www.opentk.com/node/330
	/// </summary>
	internal class Joystick
	{
		#region JOYINFOEX (helper struct)
		[StructLayout(LayoutKind.Sequential)]
		private struct JOYINFOEX
		{
			public uint Size;
			public uint Flags;
			public uint Xpos;
			public uint Ypos;
			public uint Zpos;
			public uint Rpos;
			public uint Upos;
			public uint Vpos;
			public uint Buttons;
			public uint ButtonNumber;
			public uint POV;
			private uint reserved1;
			private uint reserved2;
		}
		#endregion

		#region Constants
		internal const string WinmmLib = "Winmm.dll";
		#endregion

		#region Private
		private JOYCAPS[] gamePadCaps;
		private JOYINFOEX[] gamePadInfos;

		private int numberOfUsedDevices;
		private int joyCapsSize;
		#endregion

		#region Public
		public readonly bool[] IsConnected;
		#endregion

		#region Constructor
		public Joystick()
		{
			numberOfUsedDevices = 0;
			joyCapsSize = Marshal.SizeOf(typeof(JOYCAPS));

			uint numDevices = joyGetNumDevs();
			gamePadCaps = new JOYCAPS[numDevices];
			gamePadInfos = new JOYINFOEX[numDevices];
			IsConnected = new bool[numDevices];
		}
		#endregion

		#region System Invokes
		[DllImport(WinmmLib, EntryPoint = "joyGetDevCaps")]
		private static extern uint joyGetDevCaps(uint id, out JOYCAPS caps, int cbjc);

		[DllImport(WinmmLib, EntryPoint = "joyGetNumDevs")]
		private static extern uint joyGetNumDevs();

		[DllImport(WinmmLib, EntryPoint = "joyGetPosEx")]
		private static unsafe extern uint joyGetPosEx(uint id, JOYINFOEX* info);
		#endregion

		#region Update
		public void Update()
		{
      int count = 0;
			for (uint index = 0; index < gamePadCaps.Length; index++)
			{
				IsConnected[index] =
					joyGetDevCaps(index, out gamePadCaps[index], joyCapsSize) == 0;

				if (IsConnected[index])
				{
					count++;
				}
				else
				{
					break;
				}
			}

			if (count != numberOfUsedDevices)
			{
				numberOfUsedDevices = count;
			}

			for (int index = 0; index < numberOfUsedDevices; index++)
			{
				gamePadInfos[index] = new JOYINFOEX
				{
					Size = (uint)Marshal.SizeOf(typeof(JOYINFOEX)),
					Flags = (uint)JoystickDwFlags.JOY_RETURNALL,
				};

				unsafe
				{
					fixed (JOYINFOEX* ptr = &gamePadInfos[index])
					{
						joyGetPosEx((uint)index, ptr);
					}
				}
			}
		}
		#endregion

		#region GetThumbSticks
		public void GetThumbSticks(int index, out float xAxisLeft, out float yAxisLeft,
			out float xAxisRight, out float yAxisRight)
		{
			xAxisLeft = ((gamePadInfos[index].Xpos - gamePadCaps[index].wXmin) *
				(2f / (gamePadCaps[index].wXmax - gamePadCaps[index].wXmin))) - 1f;

			yAxisLeft = ((gamePadInfos[index].Ypos - gamePadCaps[index].wYmin) *
				(2f / (gamePadCaps[index].wYmax - gamePadCaps[index].wYmin))) - 1f;

			xAxisRight = ((gamePadInfos[index].Upos - gamePadCaps[index].wUmin) *
				(2f / (gamePadCaps[index].wUmax - gamePadCaps[index].wUmin))) - 1f;

			yAxisRight = ((gamePadInfos[index].Rpos - gamePadCaps[index].wRmin) *
				(2f / (gamePadCaps[index].wRmax - gamePadCaps[index].wRmin))) - 1f;
		}
		#endregion
		
		// TODO
		#region GetTriggers
		public void GetTriggers(int index, out float triggerLeft,
			out float triggerRight)
		{
			uint zTrigger = gamePadInfos[index].Zpos;
			triggerLeft = zTrigger == 65407 ? 1 : 0;
			triggerRight = zTrigger == 127 ? 1 : 0;
		}
		#endregion

		#region GetDPad
		public void GetDPad(int index, out bool isTopDown, out bool isBottomDown,
			out bool isLeftDown, out bool isRightDown)
		{
			isTopDown = false;
			isBottomDown = false;
			isLeftDown = false;
			isRightDown = false;

			if(gamePadInfos[index].POV == 0)
			{
				isTopDown = true;
			}
			if(gamePadInfos[index].POV == 4500)
			{
				isTopDown = true;
				isRightDown = true;
			}
			if(gamePadInfos[index].POV == 9000)
			{
				isRightDown = true;
			}
			if(gamePadInfos[index].POV == 13500)
			{
				isBottomDown = true;
				isRightDown = true;
			}
			if(gamePadInfos[index].POV == 18000)
			{
				isBottomDown = true;
			}
			if(gamePadInfos[index].POV == 22500)
			{
				isBottomDown = true;
				isLeftDown = true;
			}
			if(gamePadInfos[index].POV == 27000)
			{
				isLeftDown = true;
			}
			if(gamePadInfos[index].POV == 31500)
			{
				isTopDown = true;
				isLeftDown = true;
			}
		}
		#endregion

		// TODO
		#region GetButtons
		public void GetButtons(int index, out bool isAPressed, out bool isBPressed,
			out bool isXPressed, out bool isYPressed, out bool isBackPressed,
			out bool isStartPressed, out bool isLeftShoulderPressed,
			out bool isRightShoulderPressed, out bool isLeftThumbPressed,
			out bool isRightThumbPressed, out bool isBigPressed)
		{
			//int numButtons = gamePadCaps[index].wNumButtons;

			JoystickButtons buttons = (JoystickButtons)gamePadInfos[index].Buttons;

			isAPressed = (buttons | JoystickButtons.A) == buttons;
			isBPressed = (buttons | JoystickButtons.B) == buttons;
			isXPressed = (buttons | JoystickButtons.X) == buttons;
			isYPressed = (buttons | JoystickButtons.Y) == buttons;
			isBackPressed = (buttons | JoystickButtons.Back) == buttons;
			isStartPressed = (buttons | JoystickButtons.Start) == buttons;
			isLeftShoulderPressed = (buttons | JoystickButtons.ShoulderLeft) == buttons;
			isRightShoulderPressed = (buttons | JoystickButtons.ShoulderRight) == buttons;
			isLeftThumbPressed = (buttons | JoystickButtons.ThumbStickLeft) == buttons;
			isRightThumbPressed = (buttons | JoystickButtons.ThumbStickRight) == buttons;

			// TODO
			isBigPressed = false;
		}
		#endregion
	}
}
