﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace AC.Input.Platforms.Windows
{
	/// <summary>
	/// http://msdn.microsoft.com/en-us/library/ms644968%28v=vs.85%29.aspx
	/// </summary>
	internal class MouseHook : WindowsHook
	{
		#region MOUSEHOOKSTRUCT (Helper struct)
		private struct MOUSEHOOKSTRUCT
		{
			public int X;
			public int Y;
			public IntPtr hwnd;
			public uint wHitTestCode;
			public IntPtr dwExtraInfo;
			public int MouseData;
		}
		#endregion

		#region Constants
		public const int WM_MOUSEMOVE = 0x0200;

		public const int WM_LBUTTONDOWN = 0x0201;
		public const int WM_LBUTTONUP = 0x0202;
		public const int WM_LBUTTONDBLCLK = 0x0203;

		public const int WM_RBUTTONDOWN = 0x0204;
		public const int WM_RBUTTONUP = 0x0205;
		public const int WM_RBUTTONDBLCLK = 0x0206;

		public const int WM_MBUTTONDOWN = 0x0207;
		public const int WM_MBUTTONUP = 0x0208;
		public const int WM_MBUTTONDBLCLK = 0x0209;

		public const int WM_XBUTTONDOWN = 0x020B;
		public const int WM_XBUTTONUP = 0x020C;
		public const int WM_XBUTTONDBLCLK = 0x020D;

		public const int WM_NCMOUSEMOVE = 0x00A0;
		public const int WM_NCLBUTTONDOWN = 0x00A1;
		public const int WM_NCLBUTTONUP = 0x00A2;
		public const int WM_NCLBUTTONDBLCLK = 0x00A3;
		public const int WM_NCRBUTTONDOWN = 0x00A4;
		public const int WM_NCRBUTTONUP = 0x00A5;
		public const int WM_NCRBUTTONDBLCLK = 0x00A6;
		public const int WM_NCMBUTTONDOWN = 0x00A7;
		public const int WM_NCMBUTTONUP = 0x00A8;
		public const int WM_NCMBUTTONDBLCLK = 0x00A9;
		public const int WM_NCXBUTTONDOWN = 0x00AB;
		public const int WM_NCXBUTTONUP = 0x00AC;
		public const int WM_NCXBUTTONDBLCLK = 0x00AD;

		public const int WM_MOUSEWHEEL = 0x020A;

		public const int VK_XBUTTON1 = 0x05;
		public const int VK_XBUTTON2 = 0x06;
		#endregion

		#region Public
		public int X;
		public int Y;
		public int ScrollMovement;

		public Queue<bool> LeftButton
		{
			get;
			private set;
		}

		public Queue<bool> MiddleButton
		{
			get;
			private set;
		}

		public Queue<bool> RightButton
		{
			get;
			private set;
		}

		public Queue<bool> X1Button
		{
			get;
			private set;
		}

		public Queue<bool> X2Button
		{
			get;
			private set;
		}
		#endregion

		#region Constructor
		public MouseHook()
			: base(HookType.WH_MOUSE)
		{
			LeftButton = new Queue<bool>();
			MiddleButton = new Queue<bool>();
			RightButton = new Queue<bool>();
			X1Button = new Queue<bool>();
			X2Button = new Queue<bool>();
		}
		#endregion

		#region HandleProcMessage
		protected override void HandleProcMessage(int code, IntPtr wParam, IntPtr lParam)
		{
			if (code == HC_ACTION)
			{
				MOUSEHOOKSTRUCT mouseInfo = (MOUSEHOOKSTRUCT)Marshal.PtrToStructure(
					lParam, typeof(MOUSEHOOKSTRUCT));

				X = mouseInfo.X;
				Y = mouseInfo.Y;

				RaiseMouseHookEvent(wParam, mouseInfo.MouseData);
			}
		}
		#endregion

		#region RaiseMouseHookEvent
		private void RaiseMouseHookEvent(IntPtr wParam, int mouseData)
		{
			MouseButton button = GetMessageButton(wParam, mouseData);

			switch (wParam.ToInt32())
			{
				case WM_MOUSEMOVE:
				case WM_NCMOUSEMOVE:
					break;

				case WM_MOUSEWHEEL:
					ScrollMovement = mouseData / 120;
					break;

				case WM_LBUTTONDOWN:
				case WM_RBUTTONDOWN:
				case WM_MBUTTONDOWN:
				case WM_XBUTTONDOWN:
				case WM_NCLBUTTONDOWN:
				case WM_NCRBUTTONDOWN:
				case WM_NCMBUTTONDOWN:
				case WM_NCXBUTTONDOWN:
					switch (button)
					{
						case MouseButton.Left:
							AddtoQueue(LeftButton, true);
							break;
						case MouseButton.Middle:
							AddtoQueue(MiddleButton, true);
							break;
						case MouseButton.Right:
							AddtoQueue(RightButton, true);
							break;
						case MouseButton.XButton1:
							AddtoQueue(X1Button, true);
							break;
						case MouseButton.XButton2:
							AddtoQueue(X2Button, true);
							break;
					}
					break;

				case WM_LBUTTONUP:
				case WM_RBUTTONUP:
				case WM_MBUTTONUP:
				case WM_XBUTTONUP:
				case WM_NCLBUTTONUP:
				case WM_NCRBUTTONUP:
				case WM_NCMBUTTONUP:
				case WM_NCXBUTTONUP:
					switch (button)
					{
						case MouseButton.Left:
							AddtoQueue(LeftButton, false);
							break;
						case MouseButton.Middle:
							AddtoQueue(MiddleButton, false);
							break;
						case MouseButton.Right:
							AddtoQueue(RightButton, false);
							break;
						case MouseButton.XButton1:
							AddtoQueue(X1Button, false);
							break;
						case MouseButton.XButton2:
							AddtoQueue(X2Button, false);
							break;
					}
					break;

				case WM_LBUTTONDBLCLK:
				case WM_RBUTTONDBLCLK:
				case WM_MBUTTONDBLCLK:
				case WM_XBUTTONDBLCLK:
				case WM_NCLBUTTONDBLCLK:
				case WM_NCRBUTTONDBLCLK:
				case WM_NCMBUTTONDBLCLK:
				case WM_NCXBUTTONDBLCLK:
					break;
			}
		}
		#endregion

		#region AddtoQueue
		private void AddtoQueue(Queue<bool> queue, bool newValue)
		{
			if (queue.Count == 0 ||
				queue.ToArray()[queue.Count - 1] != newValue)
			{
				queue.Enqueue(newValue);
			}
		}
		#endregion

		#region GetMessageButton
		private MouseButton GetMessageButton(IntPtr wParam, int mouseData)
		{
			switch (wParam.ToInt32())
			{
				default:
					return MouseButton.None;

				case WM_LBUTTONDBLCLK:
				case WM_LBUTTONDOWN:
				case WM_LBUTTONUP:
				case WM_NCLBUTTONDBLCLK:
				case WM_NCLBUTTONDOWN:
				case WM_NCLBUTTONUP:
					return MouseButton.Left;

				case WM_RBUTTONDBLCLK:
				case WM_RBUTTONDOWN:
				case WM_RBUTTONUP:
				case WM_NCRBUTTONDBLCLK:
				case WM_NCRBUTTONDOWN:
				case WM_NCRBUTTONUP:
					return MouseButton.Right;

				case WM_MBUTTONDBLCLK:
				case WM_MBUTTONDOWN:
				case WM_MBUTTONUP:
				case WM_NCMBUTTONDBLCLK:
				case WM_NCMBUTTONDOWN:
				case WM_NCMBUTTONUP:
					return MouseButton.Middle;

				case WM_XBUTTONDBLCLK:
				case WM_XBUTTONDOWN:
				case WM_XBUTTONUP:
				case WM_NCXBUTTONDBLCLK:
				case WM_NCXBUTTONDOWN:
				case WM_NCXBUTTONUP:
					return mouseData == 65536 ?
						MouseButton.XButton1 :
						MouseButton.XButton2;
			}
		}
		#endregion
	}
}
