﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace AC.Input.Platforms.Windows
{
	internal class WindowsContext : BaseContext
	{
		#region RECT (Helper struct)
		[StructLayout(LayoutKind.Sequential)]
		private struct RECT
		{
			public int Left;
			public int Top;
			public int Right;
			public int Bottom;
		}
		#endregion

		#region Constants
		internal const string User32Lib = "User32.dll";
		#endregion

		#region Private
		private int windowLocationX = 0;
		private int windowLocationY = 0;
		
		private MouseHook mouseHook;
		private KeyboardHook keyboardHook;
		private Joystick joystick;
		#endregion

		#region Constructor
		public WindowsContext()
		{
			mouseHook = new MouseHook();
			keyboardHook = new KeyboardHook();
			joystick = new Joystick();
		}

		~WindowsContext()
		{
			mouseHook.Dispose();
			mouseHook = null;
			keyboardHook.Dispose();
			keyboardHook = null;
			joystick = null;
		}
		#endregion

		#region System Invokes
		[DllImport(User32Lib, EntryPoint = "GetWindowRect")]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool GetWindowRect(IntPtr hWnd, ref RECT lpRect);

		[DllImport(User32Lib)]
		private static extern bool SetCursorPos(int X, int Y);
		#endregion

		#region Update
		public override void Update()
		{
			if(WindowHandle != IntPtr.Zero)
			{
				RECT windowRect = new RECT();
				GetWindowRect(WindowHandle, ref windowRect);
				windowLocationX = windowRect.Left;
				windowLocationY = windowRect.Top;
			}

			joystick.Update();
		}
		#endregion

		#region SetMouseValues
		public override void SetMouseValues(out int X, out int Y,
			out bool[] pressed, out bool[] released, out int scrollMovement)
		{
			X = Math.Max(0, mouseHook.X - windowLocationX -
				WindowHelper.Metrics.iBorderWidth);
			Y = Math.Max(0, mouseHook.Y - windowLocationY -
				WindowHelper.Metrics.iCaptionHeight - WindowHelper.Metrics.iBorderWidth);
			scrollMovement = mouseHook.ScrollMovement;
			mouseHook.ScrollMovement = 0;

			pressed = new bool[5];
			released = new bool[5];

			SetMouseButtonState(ref pressed[0], ref released[0], mouseHook.LeftButton);
			SetMouseButtonState(ref pressed[1], ref released[1], mouseHook.MiddleButton);
			SetMouseButtonState(ref pressed[2], ref released[2], mouseHook.RightButton);
			SetMouseButtonState(ref pressed[3], ref released[3], mouseHook.X1Button);
			SetMouseButtonState(ref pressed[4], ref released[4], mouseHook.X2Button);
		}
		#endregion

		#region SetMouseButtonState
		private void SetMouseButtonState(ref bool pressed, ref bool released,
			Queue<bool> queue)
		{
			if (queue.Count > 0)
			{
				if (queue.Dequeue())
				{
					pressed = true;
				}
				else
				{
					released = true;
				}
			}
		}
		#endregion

		#region SetGamePadValues
		public override void SetGamePadValues(int index,
			out float thumbStickLeftX, out float thumbStickLeftY,
			out float thumbStickRightX, out float thumbStickRightY,
			out float triggerLeft, out float triggerRight,
			ref bool[] dpadButtons, ref bool[] buttons)
		{
			// Thumbsticks
			joystick.GetThumbSticks(index, out thumbStickLeftX, out thumbStickLeftY,
				out thumbStickRightX, out thumbStickRightY);

			// Trigger
			joystick.GetTriggers(index, out triggerLeft, out triggerRight);

			// DPad
			joystick.GetDPad(index, out dpadButtons[0], out dpadButtons[1],
				out dpadButtons[2], out dpadButtons[3]);

			// Buttons
			joystick.GetButtons(index, out buttons[0], out buttons[1], out buttons[2],
				out buttons[3], out buttons[4], out buttons[5], out buttons[6],
				out buttons[7], out buttons[8], out buttons[9], out buttons[10]);
		}
		#endregion

		#region GetPressedKeys
		public override Keys[] GetPressedKeys()
		{
			Keys[] result;

			lock (keyboardHook.PressedKeys)
			{
				result = keyboardHook.PressedKeys.ToArray();
				keyboardHook.PressedKeys.Clear();
			}
			
			return result;
		}
		#endregion

		#region GetReleasedKeys
		public override Keys[] GetReleasedKeys()
		{
			Keys[] result;

			lock (keyboardHook.ReleasedKeys)
			{
				result = keyboardHook.ReleasedKeys.ToArray();
				keyboardHook.ReleasedKeys.Clear();
			}

			return result;
		}
		#endregion

		#region WindowHandleChanged
		protected override void WindowHandleChanged()
		{
			WindowHelper.UpdateMetrics();
		}
		#endregion

		#region CheckGamePadConnection
		public override bool CheckGamePadConnection(int index)
		{
			return joystick.IsConnected[index];
		}
		#endregion

		#region SetMousePosition
		public override void SetMousePosition(int x, int y)
		{
			SetCursorPos(windowLocationX + x, windowLocationY + y);
		}
		#endregion
	}
}
